@@ -18,9 +18,12 @@
     1. having some FUNCTION:
         - some parameters need to be optimized
         - some parameters may be fixed / constant
-        - if function accepts 'device' or 'devices' it should be type of DevicesParam (check pypaq.mpython.mptools),
+        - if function accepts 'device' or 'devices' it should be type of DevicesParam (check pypaq.mpython.devices),
           it will be used by hpmser to put proper device for each function call
         - returns a dict with 'score' or just a value (score)
+        There are two parameters of FUNCTION that may be used by hpmser:
+            - 'devices' (type of DevicesParam - check pypaq.mpython.devices) -> hpmser will send devices to FUNCTION
+            - 'hpmser_mode' -> will be set to True by hpmser
     2. define PSDD - dictionary with parameters to be optimized and the space to search in (check pypaq.pms.pasap.PaSpa)
     3. import hpmser function into your script, and run it with:
         - func << FUNCTION
@@ -33,6 +36,7 @@
 # - look for solution for exponential complexity of space management functions
 # - build hpmser multiserver
 # - limited number of permutations for given PSDD (not-continuous space search case)
+# - add option to search for minimum rather than maximum
 
 
 import os
@@ -41,13 +45,14 @@ import time
 from typing import Callable, Optional, List, Any
 
 from pypaq.hpmser.search_results import SRL
-from pypaq.lipytools.little_methods import stamp, prep_folder, get_params
-from pypaq.lipytools.plots import TBPoltter
+from pypaq.hpmser.helpers import _str_weights
+from pypaq.lipytools.little_methods import stamp, get_params
+from pypaq.lipytools.files import prep_folder
 from pypaq.lipytools.logger import set_logger
 from pypaq.lipytools.stats import msmx
-from pypaq.mpython.mptools import DevicesParam
-from pypaq.mpython.omp_nb import OMPRunnerNB, RunningWorker
-from pypaq.neuralmess.dev_manager import tf_devices
+from pypaq.mpython.devices import DevicesParam, get_devices
+from pypaq.mpython.omp import OMPRunner, RunningWorker
+from pypaq.torchness.base_elements import TBwr
 from pypaq.pms.config_manager import ConfigManager
 from pypaq.pms.paspa import PaSpa
 from pypaq.pms.base_types import PSDD, POINT, point_str
@@ -69,28 +74,12 @@ SAMPLING_CONFIG_UPD = {
     'prob_top':     0.6}
 
 
-# returns nice string of floats list
-def _str_weights(
-        all_w :List[float],
-        cut_above=      5,
-        float_prec=     4) -> str:
-    ws = '['
-    if cut_above < 5: cut_above = 5 # cannot be less than 5
-    if len(all_w) > cut_above:
-        for w in all_w[:3]: ws += f'{w:.{float_prec}f} '
-        ws += '.. '
-        for w in all_w[-2:]: ws += f'{w:.{float_prec}f} '
-    else:
-        for w in all_w: ws += f'{w:.{float_prec}f} '
-    return f'{ws[:-1]}]'
-
-
 # hyper-parameters searching function (based on OMP engine)
-#@timing # << temporary disabled
+#@timing <- temporary disabled
 def hpmser(
         func: Callable,                                 # function which parameters need to be optimized
         func_psdd: PSDD,                                # func PSDD, from here points {param: arg} will be sampled
-        func_const: Optional[POINT]=    None,           # func constant kwargs, will be updated with sample taken from PaSpa
+        func_const: Optional[POINT]=    None,           # func constant kwargs, will be updated with sample (point) taken from PaSpa
         devices: DevicesParam=          None,           # devices to use for search
         name: Optional[str]=            None,           # hpmser run name, for None stamp will be used
         add_stamp=                      True,           # adds short stamp to name, when name given
@@ -104,6 +93,7 @@ def hpmser(
         do_TB=                          True,           # plots with TB
         pref_axes: Optional[List[str]]= None,           # preferred axes for plot, put here a list of up to 3 params names ['param1',..]
         top_show_freq=                  20,             # how often top results summary will be printed
+        restart_exceptions_tasks=       False,          # does not restart task that raised exception
         raise_exceptions=               True,           # forces subprocesses to raise + print exceptions, independent from verbosity (raising subprocess exception does not break hpmser process)
         verb=                           1) -> SRL:
 
@@ -114,19 +104,18 @@ def hpmser(
                 self,
                 func: Callable,
                 func_const: Optional[POINT],
-                devices: DevicesParam):
+                devices: DevicesParam = None):
 
             self.func = func
             self.func_const = func_const if func_const else {}
+            self.devices = devices
 
-            # resolve device/devices param in func >> pass it to func if needed
+            # manage 'devices' & 'hpmser_mode' param in func >> set it in func if needed
             func_args = get_params(self.func)
             func_args = list(func_args['with_defaults'].keys()) + func_args['without_defaults']
-            devices_in_func = None
-            for pn in ['device','devices']:
-                if pn in func_args:
-                    devices_in_func = pn
-            if devices_in_func: func_const[devices_in_func] = devices
+            assert 'device' not in func_args, f'ERR: you should use "devices" instead of "device" in your "{self.func.__name__}()" function'
+            if 'devices' in func_args: self.func_const['devices'] = self.devices
+            if 'hpmser_mode' in func_args: self.func_const['hpmser_mode'] = True
 
         # processes given spoint, passes **kwargs
         def process(
@@ -173,10 +162,10 @@ def hpmser(
         try_to_load=    True)
     sampling_config = config_manager.get_config() # update in case of read from existing file
 
-    tbwr = TBPoltter(f'{hpmser_FD}/{name}') if do_TB else None
+    tbwr = TBwr(logdir=f'{hpmser_FD}/{name}') if do_TB else None
 
     if verb>0:
-        print(f'\n*** hpmser *** {name} started for: {func.__name__}, sampling config: {sampling_config}')
+        print(f'\n*** hpmser {name} *** started for: {func.__name__}, sampling config: {sampling_config}')
         if srl: print(f' search will continue with {len(srl)} results...')
 
     if not srl: srl = SRL(
@@ -205,23 +194,25 @@ def hpmser(
 
     scores_all = []
 
-    devices = tf_devices(devices=devices, verb=verb) # manage devices
+    devices = get_devices(devices=devices, namespace=None) # manage devices
 
     num_free_rw = len(devices)
 
-    omp = OMPRunnerNB(
+    omp = OMPRunner(
         rw_class=               HRW,
         rw_init_kwargs=         {'func': func, 'func_const':func_const},
         rw_lifetime=            1,
         devices=                devices,
         name=                   'OMPR_NB_Hpmser',
         ordered_results=        False,
-        print_exceptions=       verb>0 or raise_exceptions,
+        restart_ex_tasks=       restart_exceptions_tasks,
+        log_exceptions=         verb > 0 or raise_exceptions,
         raise_RWW_exception=    verb>1 or raise_exceptions,
         verb=                   verb-1)
 
     top_time = time.time()
     top_speed_save = []
+    print(f'\nhpmser starts search loop..\n -- id smooth [local diff_VS_est] topID:dist_to avg_distance/max_of_min_distances time')
     try:
         while True:
 
@@ -273,92 +264,95 @@ def hpmser(
 
             msg = omp.get_result(block=True) # get one result
             num_free_rw += 1
+            if type(msg) is dict: # we may receive str here (like: 'TASK #4 RAISED EXCEPTION') from omp that not restarts exceptions
 
-            msg_sample_num =    msg['sample_num']
-            msg_score =         msg['score']
-            msg_spoint =        msg['spoint']
-            msg_est_score =     msg['est_score']
-            msg_s_time =        msg['s_time']
+                msg_sample_num =    msg['sample_num']
+                msg_score =         msg['score']
+                msg_spoint =        msg['spoint']
+                msg_est_score =     msg['est_score']
+                msg_s_time =        msg['s_time']
 
-            # manage stochastic estimation without adding to the results
-            if msg_sample_num in stochastic_points:
-                stochastic_results.append(msg_score)
-                if len(stochastic_results) == stochastic_est and verb:
-                    print(f'\n*** stochastic estimation with {stochastic_est} points:')
-                    print(f'  > results: {_str_weights(stochastic_results)}')
-                    print(f'  > std_dev: {msmx(stochastic_results)["std"]:.3f}\n')
+                # manage stochastic estimation without adding to the results
+                if msg_sample_num in stochastic_points:
+                    stochastic_results.append(msg_score)
+                    if len(stochastic_results) == stochastic_est and verb:
+                        print(f'\n*** stochastic estimation with {stochastic_est} points:')
+                        print(f'  > results: {_str_weights(stochastic_results, float_prec=8)}')
+                        print(f'  > std_dev: {msmx(stochastic_results)["std"]:.8f}\n')
 
-            else:
-                sr = srl.add_result(
-                    point=  msg_spoint,
-                    score=  msg_score)
-                if verb>1: print(f' >> got result #{msg_sample_num}')
+                else:
+                    sr = srl.add_result(
+                        point=  msg_spoint,
+                        score=  msg_score)
+                    if verb>1: print(f' >> got result #{msg_sample_num}')
 
-                avg_dst = srl.get_avg_dst()
-                mom_dst = srl.get_mom_dst()
-                srl.save(folder=f'{hpmser_FD}/{name}')
+                    pf = f'.{srl.prec}f' # update precision of print
 
-                top_SR = srl.get_top_SR()
+                    avg_dst = srl.get_avg_dst()
+                    mom_dst = srl.get_mom_dst()
+                    srl.save(folder=f'{hpmser_FD}/{name}')
 
-                # gots new MAX
-                gots_new_max = False
-                if top_SR.id != curr_max_sr_id:
-                    prev_max_sr_id = curr_max_sr_id
-                    curr_max_sr_id = top_SR.id
-                    gots_new_max = True
+                    top_SR = srl.get_top_SR()
 
-                # current sr report
-                if verb>0:
-                    dif = sr.smooth_score - msg_est_score
-                    difs = f'{"+" if dif>0 else "-"}{abs(dif):.4f}'
+                    # gots new MAX
+                    gots_new_max = False
+                    if top_SR.id != curr_max_sr_id:
+                        prev_max_sr_id = curr_max_sr_id
+                        curr_max_sr_id = top_SR.id
+                        gots_new_max = True
 
-                    dist_to_max = srl.paspa.distance(top_SR.point, sr.point)
-                    time_passed = int(time.time() - msg_s_time)
+                    # current sr report
+                    if verb>0:
 
-                    srp =  f'{sr.id} {sr.smooth_score:.4f} [{sr.score:.4f} {difs}] {top_SR.id}:{dist_to_max:.3f}'
-                    srp += f'  avg/m:{avg_dst:.3f}/{mom_dst:.3f}  {time_passed}s'
-                    if new_sampling_config: srp += f'  new sampling config: {sampling_config}'
-                    print(srp)
+                        dif = sr.smooth_score - msg_est_score
+                        difs = f'{"+" if dif>0 else "-"}{abs(dif):{pf}}'
 
-                    # new MAX report
-                    if gots_new_max:
+                        dist_to_max = srl.paspa.distance(top_SR.point, sr.point)
+                        time_passed = int(time.time() - msg_s_time)
 
-                        msr = f'_newMAX: {top_SR.id} {top_SR.smooth_score:.4f} [{top_SR.score:.4f}] {point_str(top_SR.point)}\n'
-                        # OLD VERSION: msr = f'_newMAX: {top_SR.id} {top_SR.smooth_score:.4f} [{top_SR.score:.4f}] {srl.paspa.point_2str(top_SR.get_point)}\n'
+                        srp =  f'{sr.id} {sr.smooth_score:{pf}} [{sr.score:{pf}} {difs}] {top_SR.id}:{dist_to_max:.3f}'
+                        srp += f'  avg/mom:{avg_dst:.3f}/{mom_dst:.3f}  {time_passed}s'
+                        if new_sampling_config: srp += f'  new sampling config: {sampling_config}'
+                        print(srp)
 
-                        prev_sr = srl.get_SR(prev_max_sr_id)
-                        dp = srl.paspa.distance(prev_sr.point, top_SR.point) if prev_sr else 0
+                        # new MAX report
+                        if gots_new_max:
 
-                        msr += f' dst_prev:{dp:.3f}\n'
-                        for nps in NP_SMOOTH:
-                            ss_np, avd, all_sc = srl.smooth_point(top_SR, nps)
-                            msr += f'  NPS:{nps} {ss_np:.4f} [{max(all_sc):.4f}-{min(all_sc):.4f}] {avd:.3f}\n'
-                        print(msr)
+                            msr = f'_newMAX: {top_SR.id} {top_SR.smooth_score:{pf}} [{top_SR.score:{pf}}] {point_str(top_SR.point)}\n'
 
-                    if top_show_freq and len(srl) % top_show_freq == 0:
-                        speed = int((time.time()-top_time) / top_show_freq)
-                        top_time = time.time()
-                        top_speed_save.append(speed)
-                        if len(top_speed_save) > 10: top_speed_save.pop(0)
-                        diff = int(speed - (sum(top_speed_save) / len(top_speed_save)))
-                        print(f'\n ### hpmser speed: {speed} sec/task, diff: {"+" if diff >=0 else "-"}{abs(diff)} sec')
+                            prev_sr = srl.get_SR(prev_max_sr_id)
+                            dp = srl.paspa.distance(prev_sr.point, top_SR.point) if prev_sr else 0
 
-                        print(srl.nice_str(n_top=4, top_nps=NP_SMOOTH, all_nps=None))
+                            msr += f' dst_prev:{dp:.3f}\n'
+                            for nps in NP_SMOOTH:
+                                ss_np, avd, all_sc = srl.smooth_point(top_SR, nps)
+                                msr += f'  NPS:{nps} {ss_np:{pf}} [{max(all_sc):{pf}}-{min(all_sc):{pf}}] {avd:.3f}\n'
+                            print(msr)
 
-                if tbwr:
-                    scores_all.append(sr.score)
-                    score_diff = sr.score - msg_est_score
-                    score_avg = sum(scores_all)/len(scores_all)
-                    step = len(srl)
-                    tbwr.log(avg_dst,         'hpmser/avg_dst', step)
-                    tbwr.log(score_avg,       'hpmser/score_avg', step)
-                    tbwr.log(sr.score,        'hpmser/score_current', step)
-                    tbwr.log(score_diff,      'hpmser/space_estimation_error', step)
-                    tbwr.log(abs(score_diff), 'hpmser/space_estimation_error_abs', step)
+                        if top_show_freq and len(srl) % top_show_freq == 0:
+                            speed = int((time.time()-top_time) / top_show_freq)
+                            top_time = time.time()
+                            top_speed_save.append(speed)
+                            if len(top_speed_save) > 10: top_speed_save.pop(0)
+                            diff = int(speed - (sum(top_speed_save) / len(top_speed_save)))
+                            print(f'\n ### hpmser speed: {speed} sec/task, diff: {"+" if diff >=0 else "-"}{abs(diff)} sec')
 
-                if len(srl) == n_loops:
-                    if verb>0: print(f'{n_loops} loops done!')
-                    break
+                            print(srl.nice_str(n_top=4, top_nps=NP_SMOOTH, all_nps=None))
+
+                    if tbwr:
+                        scores_all.append(sr.score)
+                        score_diff = sr.score - msg_est_score
+                        score_avg = sum(scores_all)/len(scores_all)
+                        step = len(srl)
+                        tbwr.add(avg_dst,         'hpmser/avg_dst',                    step)
+                        tbwr.add(score_avg,       'hpmser/score_avg',                  step)
+                        tbwr.add(sr.score,        'hpmser/score_current',              step)
+                        tbwr.add(score_diff,      'hpmser/space_estimation_error',     step)
+                        tbwr.add(abs(score_diff), 'hpmser/space_estimation_error_abs', step)
+
+                    if len(srl) == n_loops:
+                        if verb>0: print(f'{n_loops} loops done!')
+                        break
 
     except KeyboardInterrupt:
         if verb>0: print(' > hpmser_GX KeyboardInterrupt-ed..')
