@@ -4,79 +4,59 @@ import os
 import pandas as pd
 import plotly.express as px
 import scipy
-from typing import List
+from typing import List, Optional
 
-from pypaq.lipytools.little_methods import prep_folder
+from pypaq.lipytools.files import prep_folder
 from pypaq.lipytools.stats import stats_pd, msmx
-from pypaq.neuralmess.get_tf import tf
 
 
-# tensorboard plotter
-class TBPoltter:
-
-    def __init__(
-            self,
-            folder: str):
-
-        self.__writer = tf.summary.FileWriter(
-            logdir=         folder,
-            flush_secs=     10)
-
-    def log(
-            self,
-            value,
-            tag: str,
-            step: int):
-
-        vsumm = tf.Summary(value=[tf.Summary.Value(tag=tag, simple_value=value)])
-        self.__writer.add_summary(vsumm, step)
-
 
 def histogram(
         val_list: list or np.array,
-        name=               'values',
-        rem_nstd: float=    0.0,    # removes values out of N*stddev
-        msmx_stats=         True,   # prints minimal stats
-        pandas_stats=       False,  # prints pandas extended stats
-        density=            True,
-        bins: int or None=  None,   # automatic for None
-        save_FD :str=       None):
+        name=                   'values',
+        rem_nstd: float=        0.0,    # removes values out of N*stddev
+        msmx_stats=             True,   # prints minimal stats
+        pandas_stats=           False,  # prints pandas extended stats
+        density=                True,
+        bins: Optional[int]=    None,   # automatic for None
+        save_FD :str=           None):
 
-    if not val_list: print('WARNING: Can not prepare histogram for empty val_list!')
+    if type(val_list) is list and not val_list or type(val_list) is np.ndarray and not len(val_list):
+        print(f'cannot prepare histogram for empty val_list!')
+        return
+    if msmx_stats: print(f' > "{name}": {msmx(val_list)["string"]}')
+    if pandas_stats: print(f' > stats with pandas for "{name}":\n{stats_pd(val_list)}')
+
+    if rem_nstd:
+        stats = msmx(val_list)
+        std = stats['std']
+        mean = stats['mean']
+        val_list = [val for val in val_list if mean - rem_nstd * std < val < mean + rem_nstd * std]
+
+        if pandas_stats:
+            print(f'\n > after removing {rem_nstd} stddev:')
+            print(stats_pd(val_list))
+
+    if not bins:
+        bins = len(set(val_list))
+        if bins>50: bins = 50
+
+    plt.clf()
+    n, x, _ = plt.hist(val_list, label=name, density=density, bins=bins, alpha=0.5)
+
+    # try build density, for some val_list it is not possible
+    try:
+        density = scipy.stats.gaussian_kde(val_list)
+        plt.plot(x, density(x))
+    except: pass
+
+    plt.legend(loc='upper right')
+    plt.grid(True)
+    if save_FD:
+        if not os.path.isdir(save_FD): os.makedirs(save_FD, exist_ok=True)
+        plt.savefig(f'{save_FD}/{name}.png')
     else:
-        if msmx_stats: print(f' > "{name}": {msmx(val_list)["string"]}')
-        if pandas_stats: print(f' > stats with pandas for "{name}":\n{stats_pd(val_list)}')
-
-        if rem_nstd:
-            stats = msmx(val_list)
-            std = stats['std']
-            mean = stats['mean']
-            val_list = [val for val in val_list if mean - rem_nstd * std < val < mean + rem_nstd * std]
-
-            if pandas_stats:
-                print(f'\n > after removing {rem_nstd} stddev:')
-                print(stats_pd(val_list))
-
-        if not bins:
-            bins = len(set(val_list))
-            if bins>50: bins = 50
-
-        plt.clf()
-        n, x, _ = plt.hist(val_list, label=name, density=density, bins=bins, alpha=0.5)
-
-        # try build density, for some val_list it is not possible
-        try:
-            density = scipy.stats.gaussian_kde(val_list)
-            plt.plot(x, density(x))
-        except: pass
-
-        plt.legend(loc='upper right')
-        plt.grid(True)
-        if save_FD:
-            if not os.path.isdir(save_FD): os.makedirs(save_FD, exist_ok=True)
-            plt.savefig(f'{save_FD}/{name}.png')
-        else:
-            plt.show()
+        plt.show()
 
 
 def two_dim(
