@@ -18,9 +18,9 @@
 """
 
 from copy import deepcopy
-import itertools
-from typing import List, Optional
+from typing import List, Optional, Union, Dict
 
+from pypaq.lipytools.pylogger import get_pylogger, get_hi_child
 from pypaq.textools.text_metrics import lev_dist
 from pypaq.pms.base_types import POINT, PSDD
 from pypaq.pms.paspa import PaSpa
@@ -28,6 +28,9 @@ from pypaq.pms.paspa import PaSpa
 
 class Subscriptable:
 
+    def __init__(self, logger=None):
+        self.__log = logger or get_pylogger()
+
     def __getitem__(self, key):
         return getattr(self, key)
 
@@ -60,49 +63,36 @@ class Subscriptable:
         for key in dct:
             self.__setitem__(key, dct[key])
 
-    # checks for params similarity, returns True if already got similar
+    # checks for self.params + params for similarity, returns True if already got similar
     def check_params_sim(
             self,
-            params :dict or list,
-            lev_dist_diff: int=     1):
+            params: Optional[Union[Dict,List]]= None,
+            lev_dist_diff: int=                 1):
 
-        found_any = False
-
-        # look for params not in self.keys
+        if params is None: params = []
         paramsL = params if type(params) is list else list(params.keys())
         self_paramsL = list(self.get_point().keys())
         diff_paramsL = [par for par in paramsL if par not in self_paramsL]
 
-        # prepare dictionary of lists of lowercased underscore splits of params not in self.keys
-        diff_paramsD = {}
-        for par in diff_paramsL:
-            split = par.split('_')
-            split_lower = [el.lower() for el in split]
-            perm = list(itertools.permutations(split_lower))
-            diff_paramsD[par] = [''.join(el) for el in perm]
+        all_params = sorted(list(set([p.lower() for p in self_paramsL + diff_paramsL])))
 
-        self_paramsD = {par: par.replace('_','').lower() for par in self_paramsL} # self params lowercased with removed underscores
+        #found_any = False
+        found = set()
+        for pa in all_params:
+            for pb in all_params:
+                if pa != pb:
+                    levD = lev_dist(pa,pb)
+                    if levD <= lev_dist_diff:
+                        found.add(tuple(sorted([pa, pb])))
 
-        for p_key in diff_paramsD:
-            for s_key in self_paramsD:
-                sim_keys = False
-                s_par = self_paramsD[s_key]
-                for p_par in diff_paramsD[p_key]:
-                    if len(s_par) > lev_dist_diff and len(p_par) > lev_dist_diff:
-                        levD = lev_dist(p_par,s_par)
-                        if levD <= lev_dist_diff: sim_keys = True
-                        if s_par in p_par or p_par in s_par:
-                            sim_keys = True
-                if sim_keys:
-                    if not found_any:
-                        print('\nSubscriptable was asked to check for params similarity and found:')
-                        found_any = True
-                    print(f'WARNING: keys \'{p_key}\' and \'{s_key}\' are too CLOSE !!!')
+        if found:
+            self.__log.warning('Subscriptable was asked to check for params similarity and found:')
+            for pa,pb in found: self.__log.warning(f'> params \'{pa}\' and \'{pb}\' are too CLOSE !!!')
 
-        return found_any
+        return len(found) > 0
 
     def __str__(self):
-        s = f'\n(Subscriptable):'
+        s = f'\n(Subscriptable):\n'
         pms = self.get_point()
         for k in sorted(pms.keys()):
             p = f'param: {k}'
@@ -118,17 +108,20 @@ class SubGX(Subscriptable):
             name: str,
             family: Optional[str]=  None, # family of GXable
             psdd: Optional[PSDD]=   None, # PSDD of GXable
-            verb=                   0,
+            logger=                 None,
             **kwargs):
-        self.verb = verb
+
+        self.__log = logger or get_pylogger()
+
+        Subscriptable.__init__(self, logger=get_hi_child(self.__log))
+
         self.name = name
         self.family = family
         self.psdd = psdd or {}
         self.update(kwargs)
         # INFO: all keys of self.psdd should be present in self
-        #  It is not checked now (while init) since self may be updated even after init,
-        #  it is checked while GX
-        if self.verb>0: print(f'\n *** Subscriptable *** name: {self.name} initialized, family: {self.family}, psdd: {self.psdd}')
+        #  - it is not checked now (while init) since self may be updated even after init, IT IS checked while GX
+        self.__log.info(f'*** Subscriptable *** name: {self.name} initialized, family: {self.family}, psdd: {self.psdd}')
 
     # returns self POINT limited to axes included in self.psdd
     def get_gxable_point(self) -> POINT:
@@ -158,8 +151,11 @@ class SubGX(Subscriptable):
             parent_main: "SubGX",
             parent_scnd: Optional["SubGX"]= None,
             name_child: Optional[str]=      None,
-            **kwargs                                # here PaSpa.sample_point_GX params may be given
-    ) -> POINT:
+            prob_mix=                       0.5,
+            prob_noise=                     0.3,
+            noise_scale=                    0.1,
+            prob_axis=                      0.1,
+            prob_diff_axis=                 0.3) -> POINT:
 
         assert SubGX.families_compatible(parent_main, parent_scnd), 'ERR: families not compatible'
 
@@ -182,9 +178,13 @@ class SubGX(Subscriptable):
         dna_scnd_psdd = {k: dna_scnd[k] for k in psdd_merged} if dna_scnd is not None else None
 
         point_gx = paspa_merged.sample_point_GX(
-            point_main= dna_main_psdd,
-            point_scnd= dna_scnd_psdd,
-            **kwargs)
+            point_main=     dna_main_psdd,
+            point_scnd=     dna_scnd_psdd,
+            prob_mix=       prob_mix,
+            prob_noise=     prob_noise,
+            noise_scale=    noise_scale,
+            prob_axis=      prob_axis,
+            prob_diff_axis= prob_diff_axis)
 
         # dna_child is based on parent_main with updated values of point_gx
         dna_child: POINT = deepcopy(dna_main)
